/***************************************************************************
 *
 * Copyright (c) 1999 BalaBit Computing
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Inspired by nsyslog, originally written by Darren Reed.
 *
 * $Id: afsocket.c,v 1.1 1999/04/11 19:42:53 bazsi Exp $
 *
 ***************************************************************************/

#include "afsocket.h"
#include "format.h"
#include "cfgfile.h"
#include "pkt_buffer.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>
#include <errno.h>

#include "afsocket.c.x"

int open_unix_socket(struct sockaddr_un *su, int proto, int *sockfd)
{
	int socktype = (proto == AFUNIX_PROTO_DGRAM ? SOCK_DGRAM : SOCK_STREAM);
	int fd;
	
	fd = socket(AF_UNIX, socktype, 0);
	if (fd != -1) {
		if (su) {
			unlink(su->sun_path);
			errno = 0;
			if (bind(fd, (struct sockaddr *) su, sizeof(struct sockaddr_un)) == -1) {
				werror("cannot bind af_unix socket to %z (%z)\n", su->sun_path, strerror(errno));
				close(fd);
				return 0;
			}
			chmod(su->sun_path, 0666);
		}
		*sockfd = fd;
		return 1;
	}
	/* indicate failure */
	return 0;
}

int open_inet_socket(struct sockaddr_in *sin, int proto, int *sockfd)
{
	int socktype = (proto == AFINET_PROTO_DGRAM ? SOCK_DGRAM : SOCK_STREAM);
	int fd;
	
	fd = socket(AF_INET, socktype, 0);
	if (fd != -1) {
		if (sin) {
			int tmp = 1;
			setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp));
			if (bind(fd, (struct sockaddr *) sin, sizeof(struct sockaddr_in)) == -1) {
				werror("cannot bind af_inet socket to %I:%i (%z)\n", sin->sin_addr, ntohs(sin->sin_port), strerror(errno));
				close(fd);
				return 0;
			}
		}
		*sockfd = fd;
		return 1;
	}
	/* indicate failure */
	return 0;
}

/* CLASS:
     (class
       (name afsocket_source_connection)
       (super log_handler_pipe)
       (vars
         (backend object io_backend)
         (fd simple int)
         (src object io_fd)))
*/

static int do_init_afsocket_source_connection(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afsocket_source_connection, self, c);
	
	self->src = io_read(self->backend, self->fd, 
			make_read_line(make_log_reader(self->super.next)),
			NULL);
	REMEMBER_RESOURCE(cfg->resources, &self->src->super.super);
	return 0;
}

struct afsocket_source_connection *make_afsocket_source_connection(int fd, struct io_backend *backend, struct log_handler *next)
{
	NEW(afsocket_source_connection, self);

	self->super.super.init = do_init_afsocket_source_connection;
	self->super.next = next;
	self->backend = backend;
	self->fd = fd;
	return self;
}

/* CLASS:
     (class
       (name afsocket_accept_callback)
       (super fd_listen_callback)
       (vars
         (cfg object syslog_conf)
         (backend object io_backend)
	 (next object log_handler)))
*/

static int do_accept_unix_connection(struct fd_listen_callback *c, int fd, size_t salen, struct sockaddr *sa)
{
	CAST(afsocket_accept_callback, self, c);
	struct afsocket_source_connection *conn;
	
	conn = make_afsocket_source_connection(fd, self->backend, self->next);
	return LOG_HANDLER_INIT(conn, self->cfg);
}

struct fd_listen_callback *make_afsocket_accept_callback(struct io_backend *backend, struct log_handler *next, struct syslog_conf *cfg)
{
	NEW(afsocket_accept_callback, self);
	
	self->super.f = do_accept_unix_connection;
	self->backend = backend;
	self->next = next;
	self->cfg = cfg;
	return &self->super;
}

/* CLASS:
     (class
       (name afunix_source)
       (super log_source_driver)
       (vars
         (backend object io_backend)
         (src object io_fd)
         (listen object listen_fd)
	 (saddr simple "struct sockaddr_un")
         (proto simple int)))
*/

int do_init_afunix_source(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afunix_source, self, c);
	int fd;
	
	if (open_unix_socket(&self->saddr, self->proto, &fd)) {
		switch (self->proto) {
		case AFUNIX_PROTO_STREAM:
			if ((self->listen = io_listen(self->backend, fd, 
						      make_afsocket_accept_callback(self->backend, self->super.super.next, cfg)))) {
				REMEMBER_RESOURCE(cfg->resources, &self->listen->super.super);
				return ST_OK | ST_GOON;
			}

			break;
		case AFUNIX_PROTO_DGRAM:
			self->src = io_read(self->backend, fd, 
				make_read_line(make_log_reader((struct log_handler *) self->super.super.next)), 
				NULL);
			REMEMBER_RESOURCE(cfg->resources, &self->src->super.super);
			return ST_OK | ST_GOON;
		}
	}
	return ST_FAIL | ST_QUIT;
}

struct log_source_driver *make_afunix_source(const char *name, int proto, struct io_backend *backend)
{
        NEW(afunix_source, self);

        self->super.super.super.init = do_init_afunix_source;
        self->backend = backend;

	self->saddr.sun_family = AF_UNIX;
	strncpy(self->saddr.sun_path, name, sizeof(self->saddr.sun_path));

        self->proto = proto;
        return &self->super;
}

/* CLASS:
     (class
       (name afinet_source)
       (super log_source_driver)
       (vars
         (backend object io_backend)
         (src object io_fd)
         (listen object listen_fd)
	 (saddr simple "struct sockaddr_in")
         (proto simple int)))
*/

int do_init_afinet_source(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afinet_source, self, c);
	int fd;

	if (open_inet_socket(&self->saddr, self->proto, &fd)) {
		switch (self->proto) {
		case AFINET_PROTO_STREAM:
			if ((self->listen = io_listen(self->backend, fd, 
						      make_afsocket_accept_callback(self->backend, self->super.super.next, cfg)))) {
				REMEMBER_RESOURCE(cfg->resources, &self->listen->super.super);
				return ST_OK | ST_GOON;
			}

			break;
		case AFINET_PROTO_DGRAM:
			self->src = io_read(self->backend, fd, 
				make_read_line(make_log_reader((struct log_handler *) self->super.super.next)), 
				NULL);
			REMEMBER_RESOURCE(cfg->resources, &self->src->super.super);
			return ST_OK | ST_GOON;
		}
	}
	return ST_FAIL | ST_QUIT;
}

struct log_source_driver *make_afinet_source(struct in_addr in, unsigned short port, int proto, struct io_backend *backend)
{
        NEW(afinet_source, self);

        self->super.super.super.init = do_init_afinet_source;
        self->backend = backend;

	self->saddr.sin_family = AF_INET;
	self->saddr.sin_addr = in;
	self->saddr.sin_port = htons(port);
        self->proto = proto;
        return &self->super;
}

/* CLASS:
     (class
       (name afinet_dest)
       (super log_dest_driver)
       (vars
         (name string)
         (dest object io_fd)
	 (daddr simple "struct sockaddr_in")
	 (proto simple int)
         (backend object io_backend)))
*/

static int do_init_afinet_dest(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afinet_dest, self, c);
	int fd;
	
	if (open_inet_socket(NULL, self->proto, &fd)) {

		if (connect(fd, (struct sockaddr *) &self->daddr, 
			    sizeof(self->daddr)) != -1) {
			self->dest = io_write(self->backend, fd, 
					      make_pkt_buffer(LOG_FIFO_SIZE),
					      NULL);
			self->super.writer = 
				(struct abstract_write *) self->dest->buffer;
			REMEMBER_RESOURCE(cfg->resources, 
					  &self->dest->super.super);
		}
		else {
			close(fd);
		}
		return ST_OK | ST_GOON;
	}
	return ST_FAIL | ST_QUIT;
}

struct log_dest_driver *make_afinet_dest(struct in_addr dst, unsigned short port, int proto, struct io_backend *backend)
{
	NEW(afinet_dest, self);
	
	init_dest_driver(&self->super);
	
	self->super.super.init = do_init_afinet_dest;
	self->backend = backend;
	self->daddr.sin_family = AF_INET;
	self->daddr.sin_addr = dst;
	self->daddr.sin_port = htons(port);
	self->proto = proto;
	return &self->super;
}

/* CLASS:
     (class
       (name afunix_dest)
       (super log_dest_driver)
       (vars
         (name string)
         (dest object io_fd)
	 (daddr simple "struct sockaddr_un")
	 (proto simple int)
         (backend object io_backend)))
*/

static int do_init_afunix_dest(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afunix_dest, self, c);
	int fd;
	
	if (open_inet_socket(NULL, self->proto, &fd)) {

		if (connect(fd, (struct sockaddr *) &self->daddr, 
			    sizeof(self->daddr)) != -1) {
			self->dest = io_write(self->backend, fd, 
					      make_pkt_buffer(LOG_FIFO_SIZE),
					      NULL);
			self->super.writer = 
				(struct abstract_write *) self->dest->buffer;
			REMEMBER_RESOURCE(cfg->resources, 
					  &self->dest->super.super);
		}
		else {
			close(fd);
		}
		return ST_OK | ST_GOON;
	}
	return ST_FAIL | ST_QUIT;
}

struct log_dest_driver *make_afunix_dest(const char *name, int proto, struct io_backend *backend)
{
	NEW(afunix_dest, self);
	
	init_dest_driver(&self->super);
	
	self->super.super.init = do_init_afunix_dest;
	self->backend = backend;
	self->daddr.sun_family = AF_UNIX;
	strncpy(self->daddr.sun_path, name, sizeof(self->daddr.sun_path));
	self->proto = proto;
	return &self->super;
}
